-- Project: OiL - ORB in Lua: An Object Request Broker in Lua
-- Release: 0.6
-- Title  : Interface Definition Language (IDL) specifications in Lua
-- Authors: Renato Maia <maia@inf.puc-rio.br>
-- Notes  :
--   The syntax used for description of IDL specifications is strongly based
--   on the work provided by Letcia Nogeira (i.e. LuaRep), which was mainly
--   inteded to provide an user-friendly syntax. This approach may change to
--   allow better fitting into CORBA model, since the use of LuaIDL parsing
--   facilities already provides an user-friendly way to define IDL
--   specifications. However backward compatibility may be provided whenever
--   possible.


local _G = require "_G"                                                         --[[VERBOSE]] local verbose = require "oil.verbose"
local type = _G.type
local pairs = _G.pairs
local ipairs = _G.ipairs
local rawset = _G.rawset
local rawget = _G.rawget

-- backup of string package functions to avoid name crash with string IDL type
local strlib = require("string")
local match  = strlib.match
local format = strlib.format

local ArrayedSet = require "loop.collection.ArrayedSet"
local add = ArrayedSet.add
local remove = ArrayedSet.remove

local OrderedSet = require "loop.collection.OrderedSet"

local oo = require "oil.oo"
local class = oo.class
local isinstanceof = oo.isinstanceof

local assert = require "oil.assert"
local TypeCheckers = assert.TypeCheckers
local asserttype = assert.type
local illegal = assert.illegal

--------------------------------------------------------------------------------
-- Scoped definitions management -----------------------------------------------

local function setnameof(contained, name)
	contained.name = name
	local container = contained.defined_in
	if container then
		local start, default = match(container.repID, "^(IDL:.*):(%d+%.%d+)$")
		if not start then
			illegal(container.repID, "parent scope repository ID")
		end
		contained.repID = format("%s/%s:%s", start, contained.name,
		                                     contained.version or default)
	else
		contained.repID = format("IDL:%s:%s", contained.name,
		                                      contained.version or "1.0")
	end
	if contained.definitions then
		for _, contained in ipairs(contained.definitions) do
			setnameof(contained, contained.name)
		end
	end
end


local ContainerKey = {}

local Contents = class()

function Contents:__newindex(name, contained)
	if type(name) == "string" then
		contained.defined_in = self[ContainerKey]
		setnameof(contained, name)
		self:_add(contained)
	else
		rawset(self, name, contained)
	end
end

function Contents:_add(contained)
	add(self, contained)
	rawset(self, contained.name, contained)
	return contained
end

function Contents:_remove(contained)
	contained = remove(self, contained)
	if contained then
		self[contained.name] = nil
		return contained
	end
end

function Contents:_removeat(index)
	return self:_remove(self[index])
end

Contents._removebyname = Contents._removeat


local function Container(self)
	if not isinstanceof(self.definitions, Contents) then
		local contents = Contents{ [ContainerKey] = self }
		if self.definitions then
			for _, value in ipairs(self.definitions) do
				asserttype(value.name, "string", "IDL definition name")
				contents:__newindex(value.name, value)
			end
			for field, value in pairs(self.definitions) do
				if type(field) == "string" then
					contents:__newindex(field, value)
				end
			end
		end
		self.definitions = contents
	end
	return self
end


local function Contained(self)
	asserttype(self, "table", "IDL definition")
	if self.name  == nil then self.name = "" end
	if self.repID == nil then setnameof(self, self.name) end
	asserttype(self.name, "string", "IDL definition name")
	asserttype(self.repID, "string", "repository ID")
	return self
end

--------------------------------------------------------------------------------
-- Internal functions ----------------------------------------------------------

local function checkfield(field)
	asserttype(field.name, "string", "field name")
	asserttype(field.type, "idl type", "field type")
end

local function checkfields(fields)
	for _, field in ipairs(fields) do checkfield(field) end
end

--------------------------------------------------------------------------------
-- IDL element types -----------------------------------------------------------

local BasicTypes = {
	null       = true,
	void       = true,
	short      = true,
	long       = true,
	longlong   = true,
	ushort     = true,
	ulong      = true,
	ulonglong  = true,
	float      = true,
	double     = true,
	longdouble = true,
	boolean    = true,
	char       = true,
	octet      = true,
	any        = true,
	TypeCode   = true,
}

local UserTypes = {
	string    = true,
	Object    = true,
	struct    = true,
	union     = true,
	enum      = true,
	sequence  = true,
	array     = true,
	valuetype = true,
	valuebox  = true,
	typedef   = true,
	except    = true,
	interface = true,
	abstract_interface = true,
	local_interface = true,
}

local InterfaceElements = {
	attribute = true,
	operation = true,
	module    = true,
}

--------------------------------------------------------------------------------
-- Auxilary module functions ---------------------------------------------------

local function istype(object)
	return type(object) == "table" and (
	       	BasicTypes[object._type] == object or
	       	UserTypes[object._type]
	       )
end

local function isspec(object)
	return type(object) == "table" and (
	       	BasicTypes[object._type] == object or
	       	UserTypes[object._type] or
	       	InterfaceElements[object._type]
	       )
end

TypeCheckers["idl type"]    = istype
TypeCheckers["idl def."]    = isspec
TypeCheckers["^idl ([%l_|]+)$"] = function(value, name)
	if istype(value) then
		while value._type == "typedef" do
			value = value.original_type
		end
		for name in name:gmatch("[^|]+") do
			if value._type == name then
				return value
			end
		end
	end
	return nil
end


local ValueKind = {
	none = 0,
	custom = 1,
	abstract = 2,
	truncatable = 3,
}

local ValueMemberAccess = {
	private = 0,
	public = 1,
}

for list in pairs{[ValueKind]=true, [ValueMemberAccess]=true} do
	local values = {}
	for _,value in pairs(list) do values[value] = true end
	for value in pairs(values) do list[value] = value end
end


local function ibases(queue, interface)
	interface = queue:successor(interface)
	if interface then
		for _, base in ipairs(interface.base_interfaces) do
			queue:enqueue(base)
		end
		return interface
	end
end
local function basesof(self)
	local queue = OrderedSet()
	queue:enqueue(self)
	return ibases, queue
end

--------------------------------------------------------------------------------
-- Exported module table -------------------------------------------------------

local idl = {
	BasicTypes = BasicTypes,
	Container = Container,
	ContainerKey = ContainerKey,
	Contents = Contents,
	istype = istype,
	ValueKind = ValueKind,
	ValueMemberAccess = ValueMemberAccess,
	basesof = basesof,
}

--------------------------------------------------------------------------------
-- Basic types -----------------------------------------------------------------

for name in pairs(BasicTypes) do
	local basictype = {_type = name}
	idl[name] = basictype
	BasicTypes[name] = basictype
end

--------------------------------------------------------------------------------
-- User-defined type constructors ----------------------------------------------

-- Note: internal structure is optimized for un/marshalling.

idl.string = { _type = "string", maxlength = 0 }

function idl.struct(self)
	self = Container(Contained(self))
	self._type = "struct"
	if self.fields == nil then self.fields = self end
	checkfields(self.fields)
	return self
end

function idl.union(self)
	self = Container(Contained(self))
	self._type = "union"
	if self.options == nil then self.options = self end
	if self.default == nil then self.default = -1 end -- indicates no default in CDR
	
	asserttype(self.switch, "idl type", "union type discriminant")
	asserttype(self.options, "table", "union options definition")
	asserttype(self.default, "number", "union default option definition")
	
	self.selector = {} -- maps field names to labels (option selector)
	self.selection = {} -- maps labels (option selector) to options
	for index, option in ipairs(self.options) do
		checkfield(option)
		if option.label == nil then illegal(nil, "option label value") end
		self.selector[option.name] = option.label
		if index ~= self.default + 1 then
			self.selection[option.label] = option
		end
	end
	
	function self.__index(union, field)
		if rawget(union, "_switch") == self.selector[field] then
			return rawget(union, "_value")
		end
	end
	function self.__newindex(union, field, value)
		local label = self.selector[field]
		if label then
			rawset(union, "_switch", label)
			rawset(union, "_value", value)
			rawset(union, "_field", field)
		end
	end
	
	return self
end

function idl.enum(self)
	self = Contained(self)
	self._type = "enum"
	if self.enumvalues == nil then self.enumvalues = self end
	asserttype(self.enumvalues, "table", "enumeration values definition")
	
	self.labelvalue = {}
	for index, label in ipairs(self.enumvalues) do
		asserttype(label, "string", "enumeration value label")
		self.labelvalue[label] = index - 1
	end
	
	return self
end

function idl.sequence(self)
	self._type = "sequence"
	if self.maxlength   == nil then self.maxlength = 0 end
	if self.elementtype == nil then self.elementtype = self[1] end
	asserttype(self.maxlength, "number", "sequence type maximum length")
	asserttype(self.elementtype, "idl type", "sequence element type")
	return self
end

function idl.array(self)
	self._type = "array"
	asserttype(self.length, "number", "array type length")
	if self.elementtype == nil then self.elementtype = self[1] end
	asserttype(self.elementtype, "idl type", "array element type")
	return self
end

function idl.valuemember(self)
	self = Contained(self)
	self._type = "valuemember"
	checkfield(self)
	local access = ValueMemberAccess[self.access]
	if access == nil then
		assert.illegal(self.access, "value member access")
	end
	self.access = access
	return self
end

local null = idl.null
local valuemember = idl.valuemember
function idl.valuetype(self)
	self = Container(Contained(self))
	self._type = "valuetype"
	local kind = self.kind or (self.truncatable and ValueKind.truncatable)
	                       or (self.abstract and ValueKind.abstract)
	                       or ValueKind.none
	kind = ValueKind[kind]
	if kind == nil then
		assert.illegal(self.kind, "value kind")
	end
	self.kind = kind
	local base = self.base_value
	if base == nil then
		self.base_value = null
	elseif base ~= null then
		self.inherited = asserttype(base, "idl valuetype",
		                            "base in value definition", "BAD_PARAM")
	end
	if self.members == nil then self.members = self end
	local members = self.members
	local definitions = self.definitions
	for index, member in ipairs(members) do
		member = valuemember(member)
		members[index] = member
		definitions:__newindex(member.name, member)
	end
	return self
end

function idl.valuebox(self)
	self = Contained(self)
	self._type = "valuebox"
	if self.original_type == nil then self.original_type = self[1] end
	local type = self.original_type
	asserttype(type, "idl type", "type in typedef definition", "BAD_PARAM")
	local kind = type._type
	if kind == "valuetype" or kind == "valuebox" then
		assert.illegal(type, "type of value box")
	end
	return self
end

function idl.typedef(self)
	self = Contained(self)
	self._type = "typedef"
	if self.original_type == nil then self.original_type = self[1] end
	asserttype(self.original_type, "idl type", "type in typedef definition")
	return self
end

function idl.except(self)
	self = Container(Contained(self))
	self._type = "except"
	if self.members == nil then self.members = self end
	checkfields(self.members)
	return self
end

--------------------------------------------------------------------------------
-- IDL interface definitions ---------------------------------------------------

-- Note: construtor syntax is optimized for use with Interface Repository

function idl.attribute(self)
	self = Contained(self)
	self._type = "attribute"
	if self.type  == nil then self.type = self[1] end
	asserttype(self.type, "idl type", "attribute type")

	local mode = self.mode
	if mode == "ATTR_READONLY" then
		self.readonly = true
	elseif mode ~= nil and mode ~= "ATTR_NORMAL" then
		illegal(self.mode, "attribute mode")
	end
	
	return self
end

local void = idl.void
function idl.operation(self)
	self = Contained(self)
	self._type = "operation"
	
	local mode = self.mode
	if mode == "OP_ONEWAY" then
		self.oneway = true
	elseif mode ~= nil and mode ~= "OP_NORMAL" then
		illegal(self.mode, "operation mode")
	end

	self.inputs = {}
	self.outputs = {}
	if self.result and self.result ~= void then
		self.outputs[#self.outputs+1] = self.result
	end
	if self.parameters then
		for _, param in ipairs(self.parameters) do
			checkfield(param)
			if param.mode then
				asserttype(param.mode, "string", "operation parameter mode")
				if param.mode == "PARAM_IN" then
					self.inputs[#self.inputs+1] = param.type
				elseif param.mode == "PARAM_OUT" then
					self.outputs[#self.outputs+1] = param.type
				elseif param.mode == "PARAM_INOUT" then
					self.inputs[#self.inputs+1] = param.type
					self.outputs[#self.outputs+1] = param.type
				else
					illegal(param.mode, "operation parameter mode")
				end
			else
				self.inputs[#self.inputs+1] = param.type
			end
		end
	end

	if self.exceptions then
		for _, except in ipairs(self.exceptions) do
			asserttype(except, "idl except", "raised exception")
			if self.exceptions[except.repID] ~= nil then
				illegal(except.repID,
					"exception raise defintion, got duplicated repository ID")
			end
			self.exceptions[except.repID] = except
		end
	else
		self.exceptions = {}
	end

	return self
end

function idl.module(self)
	self = Container(Contained(self))
	self._type = "module"
	return self
end

function idl.interface(self)
	self = Container(Contained(self))
	self._type = "interface"
	if self.base_interfaces == nil then self.base_interfaces = self end
	asserttype(self.base_interfaces, "table", "interface base list")
	local inherited = {}
	for index, base in ipairs(self.base_interfaces) do
		inherited[index] = asserttype(base, "idl interface|abstract_interface",
		                              "interface base", "BAD_PARAM")
	end
	self.inherited = inherited
	self.hierarchy = basesof
	return self
end

function idl.abstract_interface(self)
	self = Container(Contained(self))
	self._type = "abstract_interface"
	if self.base_interfaces == nil then self.base_interfaces = self end
	asserttype(self.base_interfaces, "table", "interface base list")
	local inherited = {}
	for index, base in ipairs(self.base_interfaces) do
		inherited[index] = asserttype(base, "idl abstract_interface",
		                              "abstract interface base", "BAD_PARAM")
	end
	self.inherited = inherited
	self.hierarchy = basesof
	return self
end

function idl.local_interface(self)
	self = Container(Contained(self))
	self._type = "local_interface"
	if self.base_interfaces == nil then self.base_interfaces = self end
	asserttype(self.base_interfaces, "table", "interface base list")
	local inherited = {}
	for index, base in ipairs(self.base_interfaces) do
		inherited[index] = asserttype(base, "idl interface|local_interface",
		                              "local interface base", "BAD_PARAM")
	end
	self.inherited = inherited
	self.hierarchy = basesof
	return self
end

idl.object = idl.interface{
	repID = "IDL:omg.org/CORBA/Object:1.0",
	name = "Object",
}

local object = idl.object
function idl.Object(self)
	asserttype(self, "table", "Object type definition")
	if self.name  == nil then self.name = "" end
	if self.repID == nil then setnameof(self, self.name) end
	asserttype(self.name, "string", "Object type name")
	asserttype(self.repID, "string", "Object type repository ID")
	if self.repID == object.repID
		then self = object
		else self._type = "Object"
	end
	return self
end


--------------------------------------------------------------------------------
-- IDL types used in the implementation of OiL ---------------------------------

idl.ValueBase = idl.valuetype{
	repID = "IDL:omg.org/CORBA/ValueBase:1.0",
	name = "ValueBase",
	abstract = true,
}
idl.OctetSeq = idl.sequence{idl.octet}
idl.Version = idl.struct{{ type = idl.octet, name = "major" },
                         { type = idl.octet, name = "minor" }}


return idl
